/* 
 *    python-cad, a GL rendered, python scripted drawing engine
 *                  Copyright (C) 2002, 2003 by
 *
 *      Jaime Soffer             (jsoffer@soffernet.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef ENTITY_MARSHALL_H
#define ENTITY_MARSHALL_H 1

//class entityMarshall;

//#include "iMem.h"
#include "ierror.hpp"
#include <qgl.h> // Replace: class doesn't need QT

// Temporales: FIXME (compilar con g++3.2 provoca errores)
#include <stdlib.h> // Para exit(3)
#include <math.h> // Para sqrt(3)

#include "imath.hpp"

/**
 * Each entity is archived using one of these cards; the entity marshall
 * will request these values to draw, to evaluate boundboxes, etc.
 */
struct entityCard {

  /**
   * The OpenGL type of the entity, as defined on gl.h; 
   * can be GL_LINES, GL_LINE_STRIP, GL_TRIANGLES or GL_TRIANGLE_STRIP.
   */
  GLenum gltype;

  /**
   * A pointer to the memory area where the vertex are stored.
   */
  float* data;

  /**
   * A pointer to the memory area where the normals are stored.
   */
  float* normals;

  /**
   * A pointer to the memory area where the texture coordinates are stored.
   */
  float* texture;

  /**
   * A pointer to the memory area where the colors array is stored (unused).
   */
  float* color;

  /**
   * The amount of vertex this entity has. size*3*sizeof(float) is the length
   * of data, for example.
   */
  unsigned int size;

  /**
   * Highest y-value on this entity.
   */
  float top;

  /**
   * Lowest y-value on this entity.
   */
  float bottom;

  /**
   * Lowest x-value on this entity.
   */
  float left;

  /**
   * Highest x-value on this entity.
   */
  float right;
};

/**
 * Draws individual entities, finds the nearest entity to a point,
 * finds intersections, keeps a record of statistics.
 */
class entityMarshall {

 private:

  unsigned int points_count;
  unsigned int entities_count;
  unsigned int faces_count;
  unsigned int lines_count;

  float scale_over_viewport;

  float* active_vertex;

  bool snap;
  float snap_distance;
  float snap_vertex[3];

  float active_intersection[3];

  bool select;
  unsigned int selected_entity; // Deprecated [FIXME]
  unsigned int selected_entity_array[4]; // [TODO: Make a debug string]

  unsigned int vertex_selected_entity;
  unsigned int vertex_intersection_entity;

  bool blending;
  bool texture;
  bool lighting;

  bool duplication;

  entityCard* cards_array;
  float* colors_array; // [FIXME]
  iError err;

  /**
   * Accessor. Defines the active entity and the live [xyz] point that
   * will be considered active.
   */
  void setActiveVertex(unsigned int i, float* coordinates);

  /**
   * Draws a yellow square of side 2*arg around the currently active vertex
   */
  void drawActiveVertex(unsigned int);
  
  /**
   * Draws a yellow x of side 2*arg around the currently active intersection
   */
  void drawActiveInt(unsigned int);

 public:

  // Construction

  /**
   * Constructor. Reserves memory and Initializes values.
   */
  entityMarshall();

  /**
   * Destructor. Frees memory.
   */
  ~entityMarshall();


  // Value handling
  // Check signs and bounds [FIXME]
  /**
   * Increases the number of registered points.
   */ 
  void incrPoints(int n);

  /**
   * Called each time an entity is created. Increases the number
   * of registered entities.
   */
  void incrEntities(void);

  /**
   * Called when many entities are created at once. Increases the number
   * of registered entities.
   */
  void incrEntities(int n);

  /**
   * Increases the number of registered faces.
   */
  void incrFaces(int n);

  /**
   * Increases the number of registered lines.
   */
  void incrLines(int n);

  // May (will?) be dangerous at headless entities
  // Causes serious trouble when entities are initialized nested.
  /**
   * Keeps up to date the card of an entity when a point is added.
   */
  void incrSize(void);

  /**
   * Avoids the trouble by accepting the entity to increase size to.
   */
  void incrSize(unsigned int n);

  /**
   * Shifts through the entities to set the selected one.
   */
  void incrSelected(void);

  /**
   * Shifts through the entities to set the selected one.
   */
  void decrSelected(void);

  // Accessors

  /**
   * This replaces the double_entity_test variable in the memory pages;
   * since the policy now is that a just vreated entity has to push a point
   * to be considered valid or be overwritten with the next one, the flag
   * has to be global to all pages.
   */
  bool doubleBeginEntityTest(void);
  void doubleBeginEntityTest(bool b);

  /**
   * Returns the total of points (vertex) in memory.
   */
  int getNumPoints(void);

  /**
   * Returns the total of entities created.
   */
  int getNumEntities(void);

  /**
   * Returns the total of faces for triangle entities. Triangle strips
   * use about a point by face, while triangles use three.
   */
  int getNumFaces(void);

  /**
   * Returns the total of lines for line entities. Line strips create a 
   * line for each point pushed after the first, while lines create one
   * each two points.
   */
  int getNumLines(void);
  
  /**
   * ??? Equivalent to getNumEntities? [FIXME]
   */
  int getSize(void);

  /**
   * Returns the last intersection found.
   */
  float const * const getLastIntersection(void);

  /**
   * Called from the core each time a resize event is received.
   * It sets a variable to (width/height); that value is important 
   * to keep things like vertex markers on the appropiate size,
   * since they should remain a fixed size in pixels but GL draws them
   * scaled and relative to the zoom.
   */
  void setViewportScale(float vs);

  /**
   * Switches displaying the selection on/off 
   */
  void setSelect(bool b);

  /**
   * Switches on/off vertex snapping
   */
  void setSnap(bool b);

  /**
   * Returns true if vertex snapping is active
   */
  bool snapState(void);

  /**
   * Returns the [xyz] coordinates of the current snap
   */
  float const * const getSnapVertex(void);

  /**
   * Returns the [xyz] coordinates of the current intersection
   */
  float const * const getIntVertex(void);

  /**
   * Kludge. Adopts a vertex array from the core. Should be updated
   * each time the core cleans its memory.
   */
  void setColorsArray(float* colors);

  // Set error condition [FIXME]
  /**
   * Sets a named entity as the selected one.
   * It is mostly a python helper function. The python section gets an
   * antity as result from a function, and then chooses to set it as
   * selected (as part of the selection array) or not.
   */
  void setSelected(unsigned int i);

  /**
   * Returns the handle of the selected entity.
   */
  int getSelected(void);

  /**
   * Returns the [xyz] value of the currently selected vertex.
   */
  // Not used and a bit scary
  //float const * const getActiveVertex(void){ return active_vertex; };

  /**
   * Sets the distance that should be between the cursor and a snap to 
   * find it.
   */
  void setSnapRatio(float f);

  /**
   * Switches lighting on/off.
   * Still, needs to be applied.
   */
  void setLighting(bool b);

  /**
   * Switches blending on/off.
   */
  void setBlending(bool b);

  /**
   * Switches texturing on/off.
   */
  void setTexture(bool b);

  // Actuators

  /**
   * When called, attempts to set active the closer intersection 
   * to x, y, z
   */
  int findNearestIntersection(float x, float y, float z);
  // rename to: findNearestEntity() // ??? Why?

  /**
   * When called, tries to find the nearest entity to x, y, z and 
   * returns it. Does not set it as active.
   */
  int findByCoordinates(float x, float y, float z);

  /**
   * Returns true and sets snap_vertex
   * if a snap was found for the given coordinates
   */
  bool findClosestSnap(float x, float y, float z);

  /**
   * Clears the registry of entities. To be called when the core
   * calls clear; mostly useless, since the marshall is destroyed and
   * recreated too.
   */
  void reset(void);

  /**
   * Draws a blue bounding box around of the named entity.
   */
  void drawBoundbox(unsigned int n, float mag);

  /**
   * Draws an individual entity. Determines if lighting, texture and blending
   * need to be temporarily switched, too.
   */
  void draw(unsigned int n);

  /**
   * Helps to convert an entity to a python tuple. The arguments are the type
   * of entity and its size in points. the return is the pointer to the data.
   */
  float const * const getEntity(int i, GLenum & type, int & size);

  // Would be good to mix these three as:
  // void apply(bool b, glEnum param){ ... glEnable(param); ...}; ?
  // Stopper: the variables blending, texture, lighting are not indexed
  // In particular not against GL_ values

  /**
   * Enables or disables blending only if necessary.
   */
  void applyBlending(bool b);

  /**
   * Enables or disables texturing only if necessary.
   */
  void applyTexture(bool b);

  /**
   * Enables or disables lighting only if necessary.
   */
  void applyLighting(bool b);

  /**
   * Sets the active vertex to these coordinates. Potentially dangerous.
   * Modifies the value directly in memory.
   */
  void moveVertex(float x, float y, float z);

  /**
   * Creates a new empty entity of the given type, and initializes its card
   * to the given values.
   */
  void pushEntity(GLenum ltype,
		  float* ldata,
		  float* lnormals,
		  float* ltexture,
		  unsigned int layer);
  void updateEntity(unsigned int e,
		    GLenum ltype,
		    float* ldata,
		    float* lnormals,
		    float* ltexture,
		    unsigned int layer);

  /**
   * This function has to be called by the memory page once it's sure
   * that the new entity created will be definitive; it's needed
   * to avoid empty entities that may fill quickly the cards array.
   */
  void fixEntity(void);


  /**
   * Determines if [xyz] is inside of the boundbox of two named entities.
   * Good, fast test for entity boundbox intersection.
   * Takes the snap distance if [xyz] is on the far side.
   */
  bool pointinbox(float, float, float, unsigned int, unsigned int);

  // z unused by now
  // Dangerous at nested entity initialization
  // Possible fix = initBoundbox2d(u.int entity, fff) [FIXME]

  /**
   * Sets an initial value for a new entity's boundbox; setting this to 0
   * would make impossible to get correct values if the whole entity is far 
   * from the origin.
   */
  void initBoundbox2d(float x,float y, float);

  // Z unused by now
  // Dangerous at nested entity initialization
  /**
   * When adding a point to an entity, stretch the bounding box if needed
   * to make it fit in.
   */
  void recalculateBoundbox2d(float x, float y, float); 
};

#endif
