#include "imath.hpp"

// If an intersection exists returns true and sets 
// float* intersection to it; else returns false and 
// doesn't touch it.
// Lines (fp1,2) are [[xyz][xyz]] 6-float vectors.
// Intersection is a [xyz] 3-float vector
// OPTIMIZATIONS: Calculate only if the test point (there's always 
// a test point) snap area intersects both lines.
// Do not calculate if one of the lines is out of the viewport.
bool idorumath::findLinesIntersection(float const * const fp1,
				      float const * const fp2, 
				      float* intersection){

  // Stack [OPTIMIZE]
  const float x1 = fp1[0]; const float y1 = fp1[1];
  const float x2 = fp1[3]; const float y2 = fp1[4];
  const float x3 = fp2[0]; const float y3 = fp2[1];
  const float x4 = fp2[3]; const float y4 = fp2[4];

  // Determinant of denominator
  const float d = 
    (y4-y3)*(x2-x1) - (x4-x3)*(y2-y1);

  if(d < 0.000001 && d > -0.000001){  
    //  zero determinant, parallel lines
    return false;
  }

  // Determinant of line 1
  const float m =
    (x4-x3)*(y1-y3) - (y4-y3)*(x1-x3);

  // Determinant of line 2
  const float n =
    (x2-x1)*(y1-y3) - (y2-y1)*(x1-x3);

  if(!((0.0 < m) &&
       (m < d) &&
       (0.0 < n) &&
       (n < d))){
    // The intersection is out of the segments
    return false;
  }

  // If they intersect, find the point and store it at *intersection
  
  *(intersection + 0) = x1+(x2-x1)*(m/d);
  *(intersection + 1) = y1+(y2-y1)*(m/d);
  *(intersection + 2) = 0.f;
  
  return true;
}

float idorumath::calculateDistancePP(float* fp1, float* fp2){
  return sqrt((fp2[0] - fp1[0]) * (fp2[0] - fp1[0]) +
	      (fp2[1] - fp1[1]) * (fp2[1] - fp1[1]) + 
	      (fp2[2] - fp1[2]) * (fp2[2] - fp1[2]));
}

float idorumath::calculateSqrdDistancePP(float* fp1, float* fp2){
  return ((fp2[0] - fp1[0]) * (fp2[0] - fp1[0]) +
	  (fp2[1] - fp1[1]) * (fp2[1] - fp1[1]) + 
	  (fp2[2] - fp1[2]) * (fp2[2] - fp1[2]));
}

float idorumath::calculateDistancePP(float x, float y, float z, 
				     float * fp){
  return sqrt((x - fp[0]) * (x - fp[0]) +
	      (y - fp[1]) * (y - fp[1]) + 
	      (z - fp[2]) * (z - fp[2]));
}

float idorumath::calculateSqrdDistancePP(float x, float y, float z, 
					 float * fp){
  return ((x - fp[0]) * (x - fp[0]) +
	  (y - fp[1]) * (y - fp[1]) + 
	  (z - fp[2]) * (z - fp[2]));
}

int idorumath::calculateSqrdDistancePL(float x, float y, float z,
				       float * fp, 
				       float * ret) {
  bool retvalue = false;
  
  float sqrd_magnitude = calculateSqrdDistancePP(fp, fp + 3);
  
  // fatal error: points coincide
  // It's safe altough exppensive to return false here, 
  // because then the distance will be calculated twice
  // to the coinciding point. Read as: bad, should not happen,
  // but not a showstopper
  if(!sqrd_magnitude){ return false; }; 

  float scale = ((( x - fp[0] ) * 
		  ( fp[3] - fp[0] )) +
		 (( y - fp[1] ) * 
		  ( fp[4] - fp[1] )) +
		 (( z - fp[2] ) * 
		  ( fp[5] - fp[2] ))) /
    ( sqrd_magnitude );
  
  if(scale > 0.0f && scale < 1.0f){
    float p[3];
    
    p[0] = fp[0] + scale * ( fp[3] - fp[0] );
    p[1] = fp[1] + scale * ( fp[4] - fp[1] );
    p[2] = fp[2] + scale * ( fp[5] - fp[2] );
    
    *ret = calculateSqrdDistancePP(x, y, z, p);
    retvalue = true;
  }

  return retvalue;
}

idorumath::Quaternion::Quaternion(void){
  w = 1.f;
  x = y = z = 0.f;
}

idorumath::Quaternion::Quaternion(float nw, float nx, float ny, float nz){
  w = nw; x = nx; y = ny; z = nz;
}

idorumath::Quaternion::~Quaternion(void){
}

void idorumath::Quaternion::setData(float nw, float nx, float ny, float nz){
  w = nw; x = nx; y = ny; z = nz;
}

// A bit silly... [FIXME]
/*
void idorumath::Quaternion::toGLMatrix(float* fp){

}
*/

/*
w=w1w2 - x1x2 - y1y2 - z1z2
 x = w1x2 + x1w2 + y1z2 - z1y2
 y = w1y2 + y1w2 + z1x2 - x1z2
 z = w1z2 + z1w2 + x1y2 - y1x2 
*/

idorumath::Quaternion idorumath::Quaternion::operator*(Quaternion q){

  return( Quaternion(w*q.w - x*q.x - y*q.y - z*q.z,
		     w*q.x + x*q.w + y*q.z - z*q.y,
		     w*q.y + y*q.w + z*q.x - x*q.z,
		     w*q.z + z*q.w + x*q.y - y*q.x) );
	 
}
