/* 
 *    fandango, a GL rendered, python scripted drawing engine
 *                  Copyright (C) 2002, 2003 by
 *
 *      Jaime Soffer             (jsoffer@soffernet.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "imem.hpp"

// Must be divisible by 2 (lines) and 3 (triangles)
// and bigger than two (triangle strips)
//const unsigned int POINTS_IN_PAGE  = 4096*3;
const unsigned int POINTS_IN_PAGE  = 32*3;

iMem::iMem(GLenum page_type, entityMarshall* leader) {

  marshall = leader;
  if(marshall == NULL) err.quit("iMem: bad marshall passed");

  array_type = page_type;
  prev = NULL;

  index = 0;
  entity_index = 0;
  last_layer = 0;
  active_entity = 0;
  pointprev_x = pointprev_y = pointprev_z = 0.f;

  double_beginentity_test = false;

  first_point_in_entity = true;

  // Will be toggled only once, irreversible
  initialized = false;

  duplicated = false;

  winding = false;

  next = NULL;
  data = NULL;
  //  colors = NULL;
  normals = NULL;
  texture = NULL;

  /////////////////////
  //  Memory stuff [CLEAN]
  data = new float [POINTS_IN_PAGE * 3]; err.check(data);
  memset(data, 0, POINTS_IN_PAGE * 3 * sizeof(float));
  // If the order in the GL headers changes, this will fail badly
  if(array_type >= GL_TRIANGLES){
    normals = new float [POINTS_IN_PAGE * 3]; err.check(normals);
    memset(normals, 0, POINTS_IN_PAGE * 3 * sizeof(float));
    texture = new float [POINTS_IN_PAGE * 2]; err.check(texture);
    memset(texture, 0, POINTS_IN_PAGE * 2 * sizeof(float));
  }
  if(array_type == GL_LINE_STRIP){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  if(array_type == GL_LINES){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  if(array_type == GL_TRIANGLES){
    entity_size = new unsigned int [POINTS_IN_PAGE/3]; 
    err.check(entity_size);
    memset(entity_size, 0, POINTS_IN_PAGE/3 * sizeof(float));
    // Apparently POINTS_IN_PAGE/3 is not enough for the extreme case, 
    // segfaults at exit
    // WARNING: DOUBLE CALLS TO beginEntity() WITHOUT AN ACTUAL
    // ENTITY PUSHED WILL FILL THIS VERY FAST
  }
  if(array_type == GL_TRIANGLE_STRIP){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  //
  /////////////////////

#ifdef MEMDEBUG
  printf("iMem:iMem(%d)\n", array_type);
#endif

}

iMem::iMem(iMem* previous) {

  marshall = previous->getMarshall();

  array_type = previous->getArrayType();
  last_layer = previous->getLastLayer();
  // Buggy: should not be used with this default value
  active_entity = 0;
  prev = previous;

  pointprev_x = pointprev_y = pointprev_z = 0.f;

  index = 0;
  entity_index = 0;

  double_beginentity_test = false;

  first_point_in_entity = true;

  winding = false;

  // Will be toggled only once, irreversible
  initialized = false;

  duplicated = false;

  next = NULL;
  data = NULL;
  normals = NULL;
  texture = NULL;

  // Unused ?
  // entity_layer = NULL;

  /////////////////////
  //  Memory stuff
  data = new float [POINTS_IN_PAGE * 3]; err.check(data);
  memset(data, 0, POINTS_IN_PAGE * 3 * sizeof(float));
  // If the order in the GL headers changes, this will fail badly
  if(array_type >= GL_TRIANGLES){
    normals = new float [POINTS_IN_PAGE * 3]; err.check(normals);
    memset(normals, 0, POINTS_IN_PAGE * 3 * sizeof(float));
    texture = new float [POINTS_IN_PAGE * 2]; err.check(texture);
    memset(texture, 0, POINTS_IN_PAGE * 2 * sizeof(float));
  }
  if(array_type == GL_LINE_STRIP){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  if(array_type == GL_LINES){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  if(array_type == GL_TRIANGLES){
    entity_size = new unsigned int [POINTS_IN_PAGE/3]; 
    err.check(entity_size);
    memset(entity_size, 0, POINTS_IN_PAGE/3 * sizeof(float));
  }
  if(array_type == GL_TRIANGLE_STRIP){
    entity_size = new unsigned int [POINTS_IN_PAGE]; err.check(entity_size);
    memset(entity_size, 0, sizeof(int)*POINTS_IN_PAGE);
  }
  //
  /////////////////////

#ifdef MEMDEBUG
  printf("iMem:iMem(iMem* previous, %d)\n", array_type);
#endif
}

iMem::~iMem(void) {

  //////
  // Delete memory stuff
  // Note: if something starts to segfault in this area,
  // check that all this pointers should be initialized to NULL
  // at the constructor
  if(data != NULL) {
    delete []data; data = NULL;
  }
  if(normals != NULL){
    delete []normals; normals = NULL;
  }
  if(texture != NULL){
    delete [] texture; texture = NULL;
  }
  if(entity_size != NULL){
    delete []entity_size; entity_size = NULL;
  }
  //
  //////////
  
  if(next != NULL) {  // Delete recursively the whole list
    delete next; next = NULL;
  }

#ifdef MEMDEBUG
  printf("iMem::~iMem(%d)\n", array_type);
#endif

}

// Have to keep the real page where the point is being pushed
// Make a change in iCore to push the point on the current page, 
// not on the head. That will avoid having to cross the whole linked
// list to push the point.
void iMem::pushPoint(float x, float y, float z){

  if(index < POINTS_IN_PAGE){

    marshall->incrPoints(1);

    //// HEADER
    
    // After pushing a point any call to beginEntity is not to update
    // but to create a new one
    duplicated = false;

    if((index == 0) &&
       !initialized) {
      if(prev != NULL){
	this->beginEntity(prev->getLastLayer());
      } else {
	// If the first page is uninitialized use the default layer.
	this->beginEntity(1); // DEF_LAYER@kfandangowidget.cpp = 1
      }
    }

    // This section is a complement to beginEntity; it's called here
    // because beginEntity can be called as many times as necessary
    // and is only permanent when used.
    
    if(first_point_in_entity){
      marshall->initBoundbox2d(x, y, z);
      marshall->doubleBeginEntityTest(false);
      first_point_in_entity = false;
    } else {
      // It's going to be a problem: recalculates the current entity,
      // not the proper one in this page (FIXME)
      marshall->recalculateBoundbox2d(x, y, z);
    }

    //// END HEADER

    // For the individual lines:
    if(array_type == GL_LINES){

      data[index*3] = x;
      data[(index*3)+1] = y;
      data[(index*3)+2] = z;
      marshall->incrSize(active_entity);
      ++index;
    }  


    // For individual triangles:
    if(array_type == GL_TRIANGLES){

      // Assign the point...
      data[index*3] = x;
      data[(index*3)+1] = y;
      data[(index*3)+2] = z;
      //      printf("%d: %f, %f, %f\n", index, x, y, z);
      marshall->incrSize(active_entity);
      ++index;
      //
      // add the normal calculation here, only every three points
      //
      if(!(index % 3)){
	// Hazard :)
	// Triplicating the normal, to use it with arrays
	// and improve extensibility
	normals_from_triangles(normals + (index - 3)*3,
			       data+((index-3)*3));
	normals_from_triangles(normals + (index - 3)*3+3,
			       data+((index-3)*3));
	normals_from_triangles(normals + (index - 3)*3+6,
			       data+((index-3)*3));
      }

    }  


    // Triangle strips.
    // Pendiente:
    // Trasladar las normales al duplicar en tiras incompletas
    if(array_type == GL_TRIANGLE_STRIP){

      if(index == (POINTS_IN_PAGE - 1)){ // duplicate last two points

	data[index*3] = x;
	data[(index*3)+1] = y;
	data[(index*3)+2] = z;
	marshall->incrSize(active_entity);
	++index;
	if(next == NULL){
	  next = new iMem(this); err.check(next);
	}

	// This point and the previous should be pushed to the next page.
	// The order depends on the winding order of the last triangle.
	if(!winding){
	next->pushPoint( pointprev_x,
			 pointprev_y,
			 pointprev_z);
	next->pushPoint(x, y, z);
	} else {
	next->pushPoint(x, y, z);
	next->pushPoint( pointprev_x,
			 pointprev_y,
			 pointprev_z);
	}
      } else {
	data[index*3] = x;
	data[(index*3)+1] = y;
	data[(index*3)+2] = z;
	marshall->incrSize(active_entity);
	++index;
      }

      pointprev_x = x; pointprev_y = y; pointprev_z = z;

    }


    // For the line strips:
    if(array_type == GL_LINE_STRIP){

      // Is this the last point 
      // of the page? Duplicate the point.
      if(index == (POINTS_IN_PAGE - 1)){   

	data[index*3] = x;
	data[(index*3)+1] = y;
	data[(index*3)+2] = z;
	marshall->incrSize(active_entity);
	++index;
	if(next == NULL){
	  next = new iMem(this); err.check(next);
	}
	next->pushPoint(x, y, z);
      }  else {
	// Just push the point.
	data[index*3] = x;
	data[(index*3)+1] = y;
	data[(index*3)+2] = z;
	marshall->incrSize(active_entity);
	++index;
      }

    }
    
    //// FOOTER
    
    // Extreme hack to check that every entity has a beginning
    // Mostly useless: does something very different than intended
    /*
    if(current_entity == -1) {
      err.quit("iMem::pushPoint: Headless entity; must be initialized");
    }
    */

    if(!initialized){
      err.quit("iMem::pushPoint: Headless entity; must be initialized");
    }

    // BEGIN texmap
    
    // The values don't look very exact [FIXME]

    if(array_type >= GL_TRIANGLES){
      *(texture + 2*(index - 1)) = *(data + 3*(index - 1)) / 40.0;
      *(texture + 2*(index - 1) + 1) = *(data + 3*(index - 1) + 1) / 40.0;
    }
    
    // END texmap

    if(array_type == GL_TRIANGLE_STRIP){

      // The size of the current entity (the -1 is because
      // this is being done too late; think about fixing)
      // The only reason for the bug is that the index is pre-increased
      // in beginEntity (can't ever be 0). Apparently some hack
      // which was been replaced by the 'initialized' flag; not sure.
      // If this is exactly 2, then it has exactly three points
      // That's because it hasn't been increased yet (now it's #3)
      if( *(entity_size + (entity_index-1)) == 2 ) { 

	// What is this? Since I'm beginning a new strip, I have no data
	// about the previous vertex; I give the same normal to the 
	// first three vertex on the strip.
	normals_from_triangles( normals + (index - 1)*3,
				data+((index-3)*3));
	normals_from_triangles( normals + (index - 2)*3,
				data+((index-3)*3));
	normals_from_triangles( normals + (index - 3)*3,
				data+((index-3)*3)); 
      }
      
      // Normals for the next ones, defined by a lone point
      if( *(entity_size + (entity_index-1)) > 2 ) {
	normals_from_triangles( normals + (index-1)*3,
				data+((index-3)*3));
	if(!winding){
	  *(normals + (index-1)*3) = -(*(normals + (index-1)*3));
	  *(normals + (index-1)*3 + 1) = -(*(normals + (index-1)*3 + 1));
	  *(normals + (index-1)*3 + 2) = -(*(normals + (index-1)*3 + 2));
	}
	winding = !winding;
      }

    }
    
    // Increase this entity size
    ++(*(entity_size + (entity_index-1)));
    //kdDebug() << this << " " << entity_index << endl;

    //// END FOOTER

  } else { // Put the point on the next strip.
    if(next == NULL){
      next = new iMem(this); err.check(next);
    }

    next->pushPoint(x, y, z); 
  }
  


}

// Gets to make several checks; if the entity created begins at the end of 
// a page, send it to the next page; if it is interrupted at the end of a
// page, begin a new one on the next page.

// Substitutes beginStrip() on no-strip entities
// Warning: dangerous if the programmer tries to begin a triangle/line
// when a partial primitive changed the order; must only be
// called with indexes divisible by the element size (hazard on iCore.cc)

// Only sets the beginning of the entity and the layer to which belongs;
// the points must be pushed outside.

// This crashes: (nt-nl-l-t bug)
// beginTrianglesEntity(); line_from(); triangle()
// Internally:
// iTriangles->beginEntity(); (set values a)
// iPageStrip->beginEntity(); (set values b, overwrites a)
// iPageStrip->pushPoint(); (increase entities_count b, fixates)
// iTriangles->pushPoint(); (increase entities_count a - but is empty)
// Further calls to iTriangles (i.e. drawing) will corrupt the heap
// (see commentary in fixEntity@entitymarshal.cpp)

void iMem::beginEntity(unsigned int layer) {

  last_layer = layer;

  if(index < POINTS_IN_PAGE){ 

    // This is to allow to initialize for example a triangle entity,
    // then a lines entity, without any point pushed, and to push then
    // points in parallel without interference, even if one of them is 
    // started twice (to change parameters) without having pushed any point
    if(!duplicated){

      active_entity = marshall->getNumEntities();
      marshall->pushEntity(array_type,
			   data + (index*3),
			   normals + (index*3),
			   texture + (index*2),
			   layer);

      marshall->doubleBeginEntityTest(true);

      marshall->fixEntity(); // FIXME: old code, review
      ++entity_index;

      // This page has been initialized at least once; don't be afraid
      // to push points
      initialized = true;

      duplicated = true;
      first_point_in_entity = true;

      if(array_type == GL_TRIANGLE_STRIP){
	winding = false; // counterclockwise triangle
      }
    } else {
      marshall->updateEntity(active_entity,
			     array_type, // Some of that values should not
			     data + (index*3), // be updatable (FIXME)
			     normals + (index*3),
			     texture + (index*2),
			     layer); // Maybe the only updatable?

      // Duplicated code: FIXME
      if(array_type == GL_TRIANGLE_STRIP){
	winding = false; // counterclockwise triangle
      }
    }
  } else {
    if(next == NULL){
      next = new iMem(this); err.check(next);
    }
    next->beginEntity(layer);
  }
}
