/* 
 *    fandango, a GL rendered, python scripted drawing engine
 *                  Copyright (C) 2002, 2003 by
 *
 *      Jaime Soffer             (jsoffer@soffernet.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "kfandango.hpp"

extern KFandangoWidget* gw;

KFandango::KFandango(int argc, char** argv,
		     QWidget *parent=0, 
		     const char *name=0 ) :
  KMainWindow(parent, name){

  kdDebug() << "Kfandango::Kfandango" << endl;

  setCaption(i18n("New drawing"));
  resize(800, 600);

  qvb = new QVBox(this, "holderbox");
  setCentralWidget(qvb);

  gw = new KFandangoWidget(qvb, "glview");

  qpo = new QPyObject(argc, argv, this, "interpreter");

  command = new KLineEdit(qvb, "commandline");

  // Hook the python object to the "open" dialog
  connect(this, SIGNAL( executeCommand(const QString &) ),
	  qpo, SLOT( processExternalCommand(const QString &) ));

  // Hook the CAD with the statusbar
  connect(gw, SIGNAL( mousePosString(const QString &) ),
	  statusBar(), SLOT( message(const QString &) ));
  connect(gw, SIGNAL( printStdout(const QString &) ),
	  statusBar(), SLOT( message(const QString &) ));

  // Insert item completion here
  KCompletion *comp = command->completionObject();

  // Default completion items; organize [FIXME]
  comp->addItem("arch()");
  comp->addItem("mix(func, func, 1, 7)");
  comp->addItem("qzoom(15)");
  comp->addItem("texture(1)");
  comp->addItem("blend(1)");
  comp->addItem("clp()");
  comp->addItem("beginLinesEntity");
  comp->addItem("beginTrianglesEntity");
  comp->addItem("line 0 0 10 10");
  comp->addItem("triangle -10 0 0 0 0 10");
  command->connect(command,SIGNAL(returnPressed(const QString&)),
		   comp,SLOT(addItem(const QString&)));

  command->connect(command, SIGNAL( returnPressed(const QString &) ),
		   qpo, SLOT( processExternalCommand(const QString &) ));
  command->connect(command, SIGNAL( returnPressed(const QString &) ),
		   command, SLOT(clear()));

  connect(gw, SIGNAL( executeCommand(const QString &) ),
	  qpo, SLOT( processExternalCommand(const QString &) ));

  KStdAction::quit(this, SLOT(close()), actionCollection()); 

  KStdAction::open(this, SLOT(slotOpenFile()), actionCollection());

  (void)new KPyAction(qpo, "clear()",
		      i18n("&Clear"), 
		      "filenew",
		      //KStdAccel::key(KStdAccel::New), 
		      Qt::CTRL+Qt::Key_N,
		      0, 0,
		      actionCollection(), "clear");

  (void)new KPyAction(qpo, "cadget.get_errors()",
		      i18n("List Python &errors"), 
		      "error",
		      0,
		      0, 0,
		      actionCollection(), "errors");


  (void)new KPyAction(qpo, "reset()",
		      i18n("&Reset rotation"), 
		      "gohome",
		      Qt::CTRL+Qt::Key_R,
		      0, 0,
		      actionCollection(), "reset");
  
  QString restart_string =
    "if(os.path.isfile(os.path.expanduser('~/.fandango'))): \
 execfile(os.path.expanduser('~/.fandango'))";
  (void)new KPyAction(qpo, restart_string,
		      i18n("&Restart"), 
		      "reload",
		      KStdAccel::key(KStdAccel::Reload), // deprecated
		      0, 0,
		      actionCollection(), "reload");
  
  (void)new KPyToggleAction(qpo, 
			    "lines_on()", 
			    "fandango_input.mouseLbutton = none",
			    i18n("Draw &lines"), 
			    "polyline",
			    Qt::CTRL+Qt::Key_L,
			    "lmouse", 
			    actionCollection(), "lines");

  (void)new KPyToggleAction(qpo, 
			    "fandango_input.mouseLbutton = ctriangles",
			    "fandango_input.mouseLbutton = none",
			    i18n("Draw &triangles"), 
			    "triangle",
			    Qt::CTRL+Qt::Key_T,
			    "lmouse",
			    actionCollection(), "triangles");

  (void)new KPyToggleAction(qpo, 
			    "fandango_input.mouseLbutton = cint",
			    "fandango_input.mouseLbutton = none",
			    i18n("Find &intersections"), 
			    "intersection",
			    0,
			    "lmouse",
			    actionCollection(), "intersections");

  (void)new KPyToggleAction(qpo, 
			    "fandango_input.mouseLbutton = cs",
			    "fandango_input.mouseLbutton = none; select(0)",
			    i18n("Select entity"), 
			    "colorpicker",
			    0,
			    "lmouse",
			    actionCollection(), "select");

  (void)new KPyToggleAction(qpo, 
			    "reset(); lock(1)",
			    "lock(0)",
			    i18n("Lo&ck rotation"), 
			    "lock",
			    Qt::Key_F2,
			    "blocking",
			    actionCollection(), "lock_rot");


  toolbarAction = KStdAction::showToolbar(this, 
					  SLOT(toggleToolbar()), 
					  actionCollection());
  
  statusbarAction = KStdAction::showStatusbar(this, 
					      SLOT(toggleStatusbar()), 
					      actionCollection());

  menubarAction = KStdAction::showMenubar(this, 
					  SLOT(toggleMenubar()), 
					  actionCollection());
  
  KStdAction::keyBindings(this,
			  SLOT(slotKeyBindings()),
			  actionCollection());

  KStdAction::configureToolbars(this,
				SLOT(slotConfigureToolbars()),
				actionCollection());


  createGUI();

  kdDebug() << "Kfandango::Kfandango; done" << endl;
}

KFandango::~KFandango(){
  kdDebug() << "Kfandango::~Kfandango" << endl;
}

// Wrapper function to the python object
void KFandango::initialize(bool b){
  qpo->initialize(b);
}

void KFandango::toggleToolbar(){
  if (toolbarAction->isChecked()){
    toolBar()->show();
  } else {
    toolBar()->hide();
  }

}

void KFandango::toggleStatusbar(){

  if (statusbarAction->isChecked()){
    statusBar()->show();
  } else {
    statusBar()->hide();
  }

}

void KFandango::toggleMenubar(){

  if (menubarAction->isChecked()){
    menuBar()->show();
  } else {
    menuBar()->hide();
  }

}

void KFandango::slotKeyBindings(void){
  KKeyDialog kkd(true, this);
  kkd.configure(actionCollection());
}

// Does not use autosave settings
void KFandango::slotConfigureToolbars(void){
  KEditToolbar ket(actionCollection());
  if (ket.exec()) {
    createGUI();
  }
}

void KFandango::slotOpenFile(void){

  KURL url = KFileDialog::getOpenURL(":open-cad-file",
				     i18n("*.dxf|Drawing interchange files (*.dxf)\n*.xml|Exchange markup files (*.xml)\n*.*|All files"),
				     this,
				     i18n("Open file..."));
				     
  if(!url.isEmpty() && url.hasPath()){
    setCaption(url.path());
    // Determine if the file is an DXF or XML
    emit executeCommand("load_file(\"" + url.path() + "\")");
  }

}

#include "kfandango.moc"
