/* 
 *    fandango, a GL rendered, python scripted drawing engine
 *                  Copyright (C) 2002, 2003 by
 *
 *      Jaime Soffer             (jsoffer@soffernet.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #include <stdio.h> // TEST

#include "kfandangowidget.hpp"
#include "entitymarshall.hpp"

#include <kapp.h>
#include <kmessagebox.h>

#include <qgl.h>

#include <qlineedit.h>
#include <qlabel.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

// static int timer_interval = 10;	// timer interval (millisec)

KFandangoWidget::KFandangoWidget( QWidget *parent, const char *name )
     : QGLWidget(parent, name )
{

  kdDebug() << "KfandangoWidget::KfandangoWidget" << endl;

  // All these should eventually be consts
  NUM_LAYERS = 256; BACKGROUND = 0; DEF_LAYER = 1;
  CROSSHAIR = NUM_LAYERS - 1;
  TOTAL_DEPTH = 1000.f;
  NEAR_PLANE = 100000.f;
  MIDDLE_PLANE = (TOTAL_DEPTH/2.f) + NEAR_PLANE;
  FAR_PLANE = TOTAL_DEPTH + NEAR_PLANE;

  // Custom ordenation; use a header (FIXME)
  MAX_LIGHTS = 8;
  AMBIENT = 0;
  DIFFUSE = 4;
  SPECULAR = 8;
  POSITION = 12;

  rotation_lock = false;
  oldstylerotation = true;

  // Memory management
  tex =  new iTexture; err.check(tex);
  marshall = new entityMarshall; err.check(marshall);

  iPage = new iMem(GL_LINES, marshall); err.check(iPage);
  iPageStrip = new iMem(GL_LINE_STRIP, marshall); err.check(iPageStrip);
  iTriangles = new iMem(GL_TRIANGLES, marshall); err.check(iTriangles);
  iTriStrip = new iMem(GL_TRIANGLE_STRIP, marshall); err.check(iTriStrip);

  colors_array = new float [NUM_LAYERS * 4]; err.check(colors_array);
  memset(colors_array, 0, NUM_LAYERS*4*sizeof(float));
  marshall->setColorsArray(colors_array);   // FIXME

  lights_array = new float [MAX_LIGHTS * 16]; err.check(lights_array);
  memset(lights_array, 0, MAX_LIGHTS*16*sizeof(float));

  memset(clicks_array, 0, sizeof(float)*12);
  memset(snap_array, 0, sizeof(float)*12); // TODO: It's not being filled
  // END Memory management

  screen_ratio = (float) width() / (float) height();

  x = 0.f; y = 0.f; z = 0.f;
  center_z = -MIDDLE_PLANE;

  rotate_angle = 0.f;
  rotate_x = rotate_z = 0.f; rotate_y = 1.f;

  // New style rotation (free rotation)
  rotate_anglenewx = 0.f; rotate_anglenewy = 0.f;
  target_x = target_y = target_z = 0;

  top = 200.f; 
  bottom = -top;
  left = -(screen_ratio * (top - bottom) / 2.0);
  right = left + (top - bottom)*screen_ratio;

  center_x = (right + left)/2.f;
  center_y = (top + bottom)/2.f;

  //  startTimer( timer_interval );
  startTimer( 10 );
  setMouseTracking(true);
  setCursor(blankCursor);

  kdDebug() << "KfandangoWidget::KfandangoWidget; done" << endl;
}

KFandangoWidget::~KFandangoWidget(void) {

  kdDebug() << "KfandangoWidget::~KfandangoWidget" << endl;

  if(tex != NULL){
    delete tex; tex = NULL;
  }
  if(marshall != NULL){
    delete marshall; marshall = NULL;
  }
  if(iPage != NULL){
    delete iPage; iPage = NULL;
  }
  if(iPageStrip != NULL){
    delete iPageStrip; iPageStrip = NULL;
  }
  if(iTriangles != NULL){
    delete iTriangles; iTriangles = NULL;
  }
  if(iTriStrip != NULL){
    delete iTriStrip; iTriStrip = NULL;
  }
  if(colors_array != NULL){
    delete []colors_array; colors_array = NULL;
  }
  if(lights_array != NULL){
    delete []lights_array; lights_array = NULL;
  }

  kdDebug() << "KfandangoWidget::~KfandangoWidget; done" << endl;
}

void KFandangoWidget::fakePythonStdout(char* cp){
  QString qs = cp;



  // Slow...
  if(fontMetrics().width(qs) > width()){
    
    // Use a message box to display the full output
    // FIXME: make it optional
    if(!qs.contains('\n')){
      KMessageBox::information(this, qs);
    }

    // Cheap shot to make the algorithm not prohibitive when starting
    // with a 30k chars string
    qs.truncate(256);

    while(fontMetrics().width(qs) > width()){
      qs.truncate(qs.length() - 1);
    }
    
    qs.truncate(qs.length() - 5);
    qs.append("...");
  }

  if(!qs.contains('\n')){ 
    emit printStdout(qs);
  }
}

void KFandangoWidget::fakePythonStderr(char* cp){
  QString qs = cp;

  qs.remove('\n');

  if(!qs.isEmpty()){
    // Magic number: set the maximum as an option (FIXME)
    if(errList.count() > 25){
      errList.pop_front(); // STL compatibility function; no better way found
    }
    errList.append(qs);
  }

  // Slow...
  if(fontMetrics().width(qs) > width()){

    // Cheap shot to make the algorithm not prohibitive when starting
    // with a 30k chars string
    qs.truncate(256);

    while(fontMetrics().width(qs) > width()){
      qs.truncate(qs.length() - 1);
    }
    
    qs.truncate(qs.length() - 5);
    qs.append("...");
  }

  qs.simplifyWhiteSpace();
  if(!qs.isEmpty()){ 
    emit printStdout(qs);
  }
}

void KFandangoWidget::reportErrors(void){
  if(errList.count() > 0){
    KMessageBox::informationList(this, 
				 i18n("Python errors and warnings:"), 
				 errList);
  } else {
    KMessageBox::information(this, i18n("No Python error messages yet"));
  }
}

void KFandangoWidget::initializeGL()
{
  
  glShadeModel( GL_SMOOTH );
  glClearColor( 0.f, 0.f, 0.f, 0.0f );
  glClearDepth( 1.0f );
  glEnable( GL_DITHER );
  glDisable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);  
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glLineWidth(1);
  glPointSize(2); // for iCore::drawCrosshair()
  //  glPolygonMode(GL_BACK, GL_LINE);

  // Lighting (culprit functions, keep an open eye)
  //  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
  glDisable(GL_COLOR_MATERIAL);
  //  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

  // default color, actually never used
  // Actually used, will mess up the backside color later
  glColor3f(0.f, 0.75f, 0.f); 

  setLayerColor(BACKGROUND, 0.f, 0.f, 0.f, 0.f);
  setLayerColor(CROSSHAIR, 1.f, 1.f, 1.f, 1.f);
  setLayerColor(DEF_LAYER, 0.f, 0.75f, 0.f, 1.f);

  // FIXME: highlight color default
  setLayerColor(253, 1.f, 0.f, 0.f, 1.f);

  // Vertex arrays
  glEnableClientState(GL_VERTEX_ARRAY);
  /*
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  */
  // Do not deal with these arrays here, because
  // the first called entity may be a lines one and
  // the no existence of a normal array will bomb
  
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  

  initTexturing();

  setLight(0, AMBIENT, 0.05f, 0.05f, 0.05f, 1.f);
  setLight(0, DIFFUSE, 1.f, 1.f, 1.f, 1.f);
  setLight(0, SPECULAR, 1.f, 1.f, 1.f, 1.f);
  // POSITION required substracting the middle plane on z. 
  setLight(0, POSITION, 25.f, 25.f, 50.f - MIDDLE_PLANE, 1.f);
  initLight(0);

  glFinish();

  //  emit initialized();
  // resize();
}


void KFandangoWidget::resizeGL( int w, int h )
{
  // Recalculate, because the screen ratio may have changed
  // Without this the screen squeezes when resizing
  screen_ratio = (float) width() / (float) height();
  left = center_x - (screen_ratio * (top - bottom) / 2.0);
  right = left + (top - bottom)*screen_ratio;
  
  glViewport(0, 0, w, h); 

  resize();  
}

void KFandangoWidget::resize(){

  glMatrixMode(GL_PROJECTION);
  
  glLoadIdentity();

  glFrustum( left*(NEAR_PLANE/MIDDLE_PLANE), 
	     right*(NEAR_PLANE/MIDDLE_PLANE),
  	     bottom*(NEAR_PLANE/MIDDLE_PLANE),
	     top*(NEAR_PLANE/MIDDLE_PLANE), 
	     NEAR_PLANE, FAR_PLANE ); 
  
  glMatrixMode(GL_MODELVIEW);

  viewport_scale = ((right - left) / width());
  marshall->setViewportScale(viewport_scale); // FIXME
  center_x = (left + right)/2.0;
  center_y = (top + bottom)/2.0;
}

void KFandangoWidget::paintGL()
{

  /*
  using idorumath::Quaternion;

  Quaternion q;
  */

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  glLoadIdentity();

  // Matrix chain
  glTranslatef(center_x,
	       center_y,
	       center_z); 

  if(oldstylerotation){
    glRotatef(rotate_angle, rotate_x, rotate_y, rotate_z);
  } else {

    // This approach can't give a real "free" rotation.
    // For that, attempt to keep a cache matrix, multiply
    // it by each free rotation and lastly multiply it right
    // here using glMultMatrix. Or use quaternion products.

    // This extrange looking transformation swaps the freedom 
    // from x-axis to y-axis (toy transformation, useless)
    glRotatef(rotate_anglenewx, 
	      cos(-rotate_anglenewy*M_PI/180.0),
	      0.f, 
	      sin(-rotate_anglenewy*M_PI/180.0));

    glRotatef(rotate_anglenewy, 0.f, 1.f, 0.f);
  }

  glTranslatef(-center_x,
	       -center_y,
	       0.f);

  for(int i = 0; i < marshall->getSize(); ++i){
    marshall->draw(i);
  }
  
  // OPTIMIZE
  glDisable(GL_LIGHTING); glDisable(GL_TEXTURE_2D); glDisable(GL_BLEND);

  // These two break consistency; everything is drawn through 
  // the marshall except these
  drawCrosshair(x, y, z, 5); // x, y, z, must be filled at mouse move
  drawBlips(5);

  applyBlending(false);
  // glFinish();
}



void KFandangoWidget::drawCrosshair(float x, float y, float z, 
				    unsigned int d){

  float size = (float) d * viewport_scale;

  glLoadIdentity();
  glTranslatef(0.f, 0.f, z - MIDDLE_PLANE);

  // Crosshair
  glColor4fv(colors_array + (4*CROSSHAIR));
  glBegin(GL_LINES);
  glVertex2f(x, bottom);
  glVertex2f(x, top);
  glVertex2f(left, y);
  glVertex2f(right, y);
  glEnd();
  
  // Little square in the cross
  glEnable(GL_BLEND);
  glDepthFunc(GL_ALWAYS);
  glColor4f(1.f, 0.f, 0.f, 0.75f);
  glBegin(GL_QUADS);
  glVertex2f(x-size, y-size);
  glVertex2f(x+size, y-size);
  glVertex2f(x+size, y+size);
  glVertex2f(x-size, y+size);
  glEnd();
  
  // A dot in the crossing
  glBegin(GL_POINTS);
  glVertex2f(x, y);
  glEnd();
  glDepthFunc(GL_LESS);
  
}

void KFandangoWidget::drawBlips(unsigned int s){

  float size = (float) s * viewport_scale;

  glDepthFunc(GL_ALWAYS);
  glColor4f(1.f, 0.f, 0.f, 0.75f);
  
  for(int i = 1; i < 4; ++i){
    glBegin(GL_LINES);
    glVertex2f(clicks_array[3*i] + size, clicks_array[3*i + 1]);
    glVertex2f(clicks_array[3*i] - size, clicks_array[3*i + 1]);
    glVertex2f(clicks_array[3*i], clicks_array[3*i + 1] + size);
    glVertex2f(clicks_array[3*i], clicks_array[3*i + 1] - size);
    glEnd();
  }

  glDepthFunc(GL_LESS);

}

void KFandangoWidget::timerEvent(QTimerEvent*)
{
    updateGL();
}

void KFandangoWidget::mouseMoveEvent(QMouseEvent * e){

  QPoint p = e->pos();

  // Rotate
  if((e->state() & LeftButton) && !rotation_lock){
    if(oldstylerotation){
      rotate_angle += (p.y() - old_pos.y());
    } else {
      rotate_anglenewy += (p.x() - old_pos.x());
      rotate_anglenewx += (p.y() - old_pos.y());
    }
  }
  
  // If Ctrl is pressed, modify the Z coordinate about x mouse translation.
  if(e->state() & ControlButton){
    z += (p.x() - old_pos.x());
  }
  old_pos = e->pos();

  if(p.x() >= 0 && p.y() >= 0){ // Don't catch out of the widget
    x = left + (right - left) * ((float)p.x() / width());
    // The top-bottom inversion is performed here
    y = bottom + (top - bottom) * ((height() - (float)p.y()) / height());
  }

  // Return information to the status bar
  QString qs_angle = i18n("angle");
  QString qs;
  if(oldstylerotation){
    qs = 
      "(" + QString::number(x) + ", " + QString::number(y) + 
      ", " + QString::number(z) +  ") " 
      + qs_angle + ": " +
      QString::number(rotate_angle) + 
      " ratio: " + QString::number(screen_ratio);
  } else {
    qs = 
      "(" + QString::number(x) + ", " + QString::number(y) + 
      ", " + QString::number(z) +  ") " +
      qs_angle + ": " +
      QString::number(rotate_anglenewx) + 
      ", " + QString::number(rotate_anglenewy);
  }

  emit mousePosString(qs);
}

void KFandangoWidget::wheelEvent(QWheelEvent * e){

  // 120 = WHEEL_DELTA, not defined anywhere (FIXME)
  QString cmd = QString::number(e->delta()/120);

  z -= float(e->delta()/120.0);

  QString qs = 
    "(" + QString::number(x) + ", " + 
    QString::number(y) + ", " + QString::number(z) + ")";

  emit mousePosString(qs);
}

// This has to fill a [[xyz3][xyz2][xyz1][xyz0]] matrix with the
// latest clicked points. Only points 0-2 require to be exact.
void KFandangoWidget::mousePressEvent(QMouseEvent * e){
  
  QPoint p = e->pos();

  // Left button handler (internal; Python has its own)
  if(e->button() == LeftButton){
   
    // Shift the clicks array 
    memmove(clicks_array, clicks_array + 3, 9*sizeof(float));

    // Add the latest click
    if(p.x() >= 0 && p.y() >= 0){ // Don't catch out of the widget
      clicks_array[9] = left + (right - left) * ((float)p.x() / width());
      clicks_array[10] = 
	bottom + (top - bottom) * ((height() - (float)p.y()) / height());
      clicks_array[11] = z;
    }

    // Inform the world
    QString qs_clicked = i18n("clicked");
    QString qs = 
      qs_clicked + ": (" + QString::number(clicks_array[9]) + 
      ", " + QString::number(clicks_array[10]) +
      ")" +
      "(" + QString::number(clicks_array[6]) + 
      ", " + QString::number(clicks_array[7]) +
      ")" +
      "(" + QString::number(clicks_array[3]) + 
      ", " + QString::number(clicks_array[4]) +
      ")";

    emit mousePosString(qs);

  }

  QString cmd = "None";

  if(e->button() == LeftButton) {
    cmd = "fandango_input.mouseLbutton()";
  }

  if(e->button() == MidButton) {
    cmd = "fandango_input.mouseMbutton()";
  }

  if(e->button() == RightButton) {
    cmd = "fandango_input.mouseRbutton()";
  }

  // Emit the generic button first
  emit executeCommand("fandango_input.mousebutton()");

  // Emit the pressed button
  emit executeCommand(cmd);

}

/*
void KFandangoWidget::keyPressEvent(QKeyEvent * e){

  QString qs = "key pressed: " + QString::number(e->state()) + " plus " +
    QString::number(e->key());
  emit mousePosString(qs);
}
*/

// Stack operations

/**************  newStrip/TriStrip/LinesEntity/TrianglesEntity  *************
 * Begins a new entity. Wrapper for the memory pages.
 */
void KFandangoWidget::newStrip(float x, float y, float z, 
			       unsigned int layer){
  iPageStrip->beginEntity(layer);
  iPageStrip->pushPoint(x, y, z);
}

void KFandangoWidget::newTriStrip(float x1, float y1, float z1,
			float x2, float y2, float z2,
			unsigned int layer) {

  iTriStrip->beginEntity(layer);
  iTriStrip->pushPoint(x1, y1, z1);
  iTriStrip->pushPoint(x2, y2, z2);
}

void KFandangoWidget::newTrianglesEntity(unsigned int layer){
  iTriangles->beginEntity(layer);
}

void KFandangoWidget::newLinesEntity(unsigned int layer){
  iPage->beginEntity(layer);
}


/*************************  pushToStrip/TriStrip  ***************************
 * Adds a point to a connected entity, increasing it by a primitive.
 * Wrapper for the memory pages.
 */
void KFandangoWidget::pushToStrip(float vx, float vy, float vz){
  iPageStrip->pushPoint(vx, vy, vz);
  marshall->incrLines(1);
}

void KFandangoWidget::pushToTriStrip(float x1, float x2, float x3){
  iTriStrip->pushPoint(x1, x2, x3);
  marshall->incrFaces(1);
}

/****************************  pushLine/Triangle  ***************************
 * Adds a new primitive to memory. It still needs to be related to an 
 * entity, though. Wrapper for the memory pages.
 */

void KFandangoWidget::pushLine(float x1, float y1, float z1,
		     float x2, float y2, float z2){

  iPage->pushPoint(x1, y1, z1);
  iPage->pushPoint(x2, y2, z2);

  marshall->incrLines(1);
}

void KFandangoWidget::pushTriangle( float x1, float y1, float z1, 
			  float x2, float y2, float z2,
			  float x3, float y3, float z3){
  iTriangles->pushPoint(x1, y1, z1);
  iTriangles->pushPoint(x2, y2, z2);
  iTriangles->pushPoint(x3, y3, z3);

  marshall->incrFaces(1);
}

void KFandangoWidget::setLayerColor(unsigned int l, 
				    float r, float g, float b, float a){

  unsigned int layer;
  if(l >= NUM_LAYERS){
    layer = DEF_LAYER;
  } else {
    layer = l;
  }

  *(colors_array + (4*layer) + 0) = r;
  *(colors_array + (4*layer) + 1) = g;
  *(colors_array + (4*layer) + 2) = b;
  *(colors_array + (4*layer) + 3) = a;
}

// Accessors

void KFandangoWidget::setLight(unsigned int light, unsigned int param,
			       float r, float g, float b, float a){
  if(light >= MAX_LIGHTS){
    err.quit("iCore::setLight: A light was set out of limits");
  }

  *(lights_array + (16*light) + param + 0) = r;
  *(lights_array + (16*light) + param + 1) = g;
  *(lights_array + (16*light) + param + 2) = b;
  *(lights_array + (16*light) + param + 3) = a;
}

void KFandangoWidget::setLightColor(unsigned int light,
				    float r, float g, float b, float a){

  setLight(light, DIFFUSE, r, g, b, a);
  glLightfv(GL_LIGHT0 + light, GL_DIFFUSE, 
  	    lights_array + (16*light) + DIFFUSE);
}

void KFandangoWidget::setLightPosition(unsigned int light,
				       float x, float y, float z){
  // Messing up with the comment below
  glLoadIdentity();

  // Attempt to keep the attenuation, no matter the viewport size [UGLY]
  float ww = (right-left)*(right-left);

  float hh = (top-bottom)*(top-bottom);

  // 0.302958 es el factor para provocar atenuacin cuadrtica
  // de 1/2000 en un qzoom de 48.75
  float q = 1.0 / (0.302958 * (ww + hh)); 
  glLightf(GL_LIGHT0 + light, GL_QUADRATIC_ATTENUATION, q);
  //  printf("Quadratic attenuation: %.6f\n", q);

  setLight(light, POSITION, x, y, z - MIDDLE_PLANE, 1.0);
  glLightfv(GL_LIGHT0 + light, GL_POSITION, 
  	    lights_array + (16*light) + POSITION);

}

// Wrappers: FIXME
void KFandangoWidget::setBlending(bool b){
  marshall->setBlending(b);
}

void KFandangoWidget::applyBlending(bool b){
  marshall->applyBlending(b);
}

void KFandangoWidget::setTexture(bool b){
  marshall->setTexture(b);
}

void KFandangoWidget::setLighting(bool b){
  marshall->setLighting(b);
}

void KFandangoWidget::setSelect(bool b){
  marshall->setSelect(b);
}

void KFandangoWidget::setSnap(bool b){
  marshall->setSnap(b);
}

// true sets axial rotation; false sets free rotation
// Free rotation ignores and overwrites the chosen axis
void KFandangoWidget::setOldRotation(bool b){
  oldstylerotation = b;
}

float const * const KFandangoWidget::getSnapVertex(void){
  return marshall->getSnapVertex();
}

float const * const KFandangoWidget::getIntVertex(void){
  return marshall->getIntVertex();
}

void KFandangoWidget::setSnapRatio(float f){
  marshall->setSnapRatio(f);
}

bool KFandangoWidget::findSnap(float x, float y, float z){
  return marshall->findClosestSnap(x, y, z);
}

bool KFandangoWidget::snapState(void){
  return marshall->snapState();
}

void KFandangoWidget::findIntersection(float x, float y, float z){
  marshall->findNearestIntersection(x, y, z);
}

float KFandangoWidget::getRatio(void){
  return ( (float) width() / (float) height() );
}

float const * const KFandangoWidget::getClickQueue(void){
  return clicks_array;
}

float const * const KFandangoWidget::getSnapQueue(void){
  return snap_array;
}

unsigned int KFandangoWidget::findByCoordinates(float x, float y, float z){
  return marshall->findByCoordinates(x, y, z);
}

void KFandangoWidget::setSelected(unsigned int i){
  marshall->setSelected(i);
}

void KFandangoWidget::lock(bool b){
  rotation_lock = b;
}

unsigned int KFandangoWidget::getSelected(void){
  return marshall->getSelected();
}

float const * const KFandangoWidget::getLastIntersection(void){
  return marshall->getLastIntersection();
}

void KFandangoWidget::moveVertex(float x, float y, float z){
  marshall->moveVertex(x, y, z);
}

float const * const KFandangoWidget::getEntity(int i, 
					       GLenum & type, int & size){
  return marshall->getEntity(i, type, size);
}

void KFandangoWidget::setViewportCorners(float l, float t, float b){

  if(t > b && screen_ratio > 0){

    left = l; top = t; bottom = b;
    right = left + (top-bottom)*screen_ratio;

    resize();
  }
}

void KFandangoWidget::initLight(unsigned int light){
  // asking for trouble... exit condition [FIXME]
  if(light >= MAX_LIGHTS){
    err.quit("iCore::initLight: A light was inited out of limits");
  }

  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHT0 + light);
  glLightf(GL_LIGHT0 + light, GL_QUADRATIC_ATTENUATION, 0.0005f);

  glLightfv(GL_LIGHT0 + light, GL_AMBIENT, 
	    lights_array + (16*light) + AMBIENT);
  glLightfv(GL_LIGHT0 + light, GL_DIFFUSE, 
	    lights_array + (16*light) + DIFFUSE);
  glLightfv(GL_LIGHT0 + light, GL_SPECULAR, 
	    lights_array + (16*light) + SPECULAR);
  glLightfv(GL_LIGHT0 + light, GL_POSITION, 
	    lights_array + (16*light) + POSITION);

  // A bit of a mess...
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

  // As a rather odd note, the back color, when calling glColorMaterial,
  // will take the current color. When no glColor had been called, the 
  // back side showed white; 
  // without the next line it shows green (iWindow.cc:~151)
  glColor4f(1.f, 1.f, 1.f, 1.f);
  //  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glColorMaterial(GL_FRONT, GL_DIFFUSE);
  // control the attenuation from outside
}

void KFandangoWidget::initTexturing(void){
  tex->makeCheckerboard();
  tex->init();
}

void KFandangoWidget::setRotationAxis(float x, float y, float z){
  // Avoid dividing by 0
  if(oldstylerotation && (x || y || z)){
    float tlen = 1.0/sqrt((x*x)+(y*y)+(z*z));
    rotate_x = x*tlen; rotate_y = y*tlen; rotate_z = z*tlen;
  }
}

void KFandangoWidget::resetTransformations(void){

  rotate_angle = 0.f;
  z = 0.0f;

  rotate_anglenewx = 0.f;
  rotate_anglenewy = 0.f;
}

void KFandangoWidget::clear(void){

  // Too rough... [FIXME]
  delete iPage; iPage = NULL;
  delete iPageStrip; iPageStrip = NULL;
  delete iTriStrip; iTriStrip = NULL;
  delete iTriangles; iTriangles = NULL;

  // The marshall should be recreated too because
  // all the page pointers are now invalid
  // It has to be done before constructing the new pages,
  // or the 'marshall' passed to them will be dangling

  delete marshall; marshall = NULL;
  marshall = new entityMarshall; err.check(marshall);

  // This line is very wrong. This functionality will have to
  // be fixed, this can cause even more trouble later.
  marshall->setColorsArray(colors_array);
  
  iPage = new iMem(GL_LINES, marshall); err.check(iPage);
  iPageStrip = new iMem(GL_LINE_STRIP, marshall); err.check(iPageStrip);
  iTriStrip = new iMem(GL_TRIANGLE_STRIP, marshall); err.check(iTriStrip);
  iTriangles = new iMem(GL_TRIANGLES, marshall); err.check(iTriangles);

  resize();

  emit executeCommand((QString) "fandango_input.clearcalled()");
}

// Without this the change of context (expensive) was made every frame in
// the first versions. Haven't tested with more current Qt.
void KFandangoWidget::makeCurrent(void){
  if(context() != context()->currentContext()){
    QGLWidget::makeCurrent();
  }
}


#include "kfandangowidget.moc"
