/* 
 *    python-cad, a GL rendered, python scripted drawing engine
 *                  Copyright (C) 2002, 2003 by
 *
 *      Jaime Soffer             (jsoffer@soffernet.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef KFANDANGOWIDGET_H
#define KFANDANGOWIDGET_H

#include <qgl.h>
#include <klocale.h>

#include <kmessagebox.h>
#include <qstringlist.h>

#include "entitymarshall.hpp"
#include "imem.hpp"
#include "itexture.hpp"
#include "ierror.hpp"

/**
 * The core of the application. Most everything not dedicated is 
 * controlled from here; the redraw timer, window resizing, 
 * python interfacing, viewport transformations and mouse interface.
 */
class KFandangoWidget : public QGLWidget
{
  Q_OBJECT

private:

  /**
   * Holds a number of previous error messages
   */
  QStringList errList;

  iError err;

  iTexture* tex;
  entityMarshall* marshall;

  iMem* iPage;
  iMem* iPageStrip;
  iMem* iTriangles;
  iMem* iTriStrip;

  QPoint old_pos;

  float* colors_array; 
  float* lights_array;

  float clicks_array[12];
  float snap_array[12];
 
  // should be consts
  // Candidates to namespace
  float TOTAL_DEPTH;
  float NEAR_PLANE, MIDDLE_PLANE, FAR_PLANE; 
  unsigned int BACKGROUND, CROSSHAIR, DEF_LAYER;
  unsigned int NUM_LAYERS;
  unsigned int MAX_LIGHTS, AMBIENT, DIFFUSE, SPECULAR, POSITION;

  bool rotation_lock;
  bool oldstylerotation;

  float top, left, bottom, right;
  float screen_ratio, viewport_scale;

  float rotate_angle;
  float rotate_x, rotate_y, rotate_z;

  float rotate_anglenewx; float rotate_anglenewy;

  // These are the vector that should be rotated toward 
  // the camera, more or less
  float target_x, target_y, target_z;

  float center_x, center_y, center_z;

  float x, y, z; // cursor coordinates

public:

  /**
   * Constructor: 
   * Reserve memory, initialize, start the timer, enable mouse tracking,
   * remove the X cursor.
   */
  KFandangoWidget( QWidget *parent=0, const char *name=0 );

  /**
   * Destructor: release memory
   */
  virtual ~KFandangoWidget(void);

  /**
   * To be called when the widget changes shape. Reshapes the frustum, 
   * sets the (width/height) scale variable, sends it to the marshall
   * and recalculates the center.
   */
  void resize(void);

  /**
   * Sets the r, g, b, a color of the layer number l.
   */
  void setLayerColor(unsigned int l, float r, float g, float b, float a);

  /**
   * Wrapper function (to the marshall)
   */
  void applyBlending(bool);

  /**
   * Draws two crossed lines across the screen, a translucid square
   * and a dot at the center, meeting at the mouse position.
   */
  void drawCrosshair(float, float, float, unsigned int);

  /**
   * Draws little red crosses at the last three clicked points. 
   */
  void drawBlips(unsigned int);

  /**
   * Emits a signal that sends a QString to another widget.
   * It replaces the sys.stdout method of Python, and all the standard
   * output is going to be redirected this way.
   */
  void fakePythonStdout(char*);

  /**
   * Emits a signal that sends a QString to another widget.
   * It replaces the sys.stderr method of Python, and all the standard
   * output is going to be redirected this way.
   */
  void fakePythonStderr(char*);

  /**
   * When called, produce an informative messagebox with the latest output
   * from stderr. FIXME: make it something more useful than a KMessageBox
   */
  void reportErrors(void);

  // Initialization

  /**
   * Calls the needed GL commands to set up texture mapping.
   */
  void initTexturing(void);

  /**
   * Calls the needed GL commands to set up lighting.
   */
  void initLight(unsigned int);

  // Accessors
  void setLight(unsigned int, unsigned int, float, float, float, float);
  void setLightColor(unsigned int, float, float, float, float);
  void setLightPosition(unsigned int, float, float, float);
  void setBlending(bool);
  void setTexture(bool);
  void setLighting(bool);
  void setSelect(bool);
  void setSnap(bool);
  void setSnapRatio(float);
  void setSelected(unsigned int);

  /**
   * If set true, prevents the mouse from rotating the view
   */
  void lock(bool);

  /**
   * Takes the desired left, top and bottom positions, and recalculates
   * the right according to the window ratio; then resizes the viewport,
   * so these values coincide with the borders of the widget.
   */
  void setViewportCorners(float, float, float);

  /**
   * If not locked, the view rotates by holding the left mouse button 
   * and rolling up-down. The center of rotation is the center of the screen,
   * and the rotation axis is set here in x, y, z coordinates.
   */
  void setRotationAxis(float, float, float);

  void setOldRotation(bool b);

  /**
   * Returns the handler of the currently selected entity. Doesn't matter if
   * select is active or not.
   */
  unsigned int getSelected(void);

  /**
   * Wrapper function (to the marshall)
   */
  float const * const getLastIntersection(void);

  /**
   * Wrapper function (to the marshall)
   */
  float const * const getSnapVertex(void);

  /**
   * Wrapper function (to the marshall)
   */
  float const * const getIntVertex(void);

  /**
   * Wrapper function (to the marshall)
   */
  float const * const getEntity(int i, GLenum & type, int & size);

  /**
   * Returns a [[xyz2][xyz1][xyz]] array of the last clicked points, in
   * age order, the older first.
   */
  float const * const getClickQueue(void);

  /**
   * Unused right now. It should work the following way: the user clicks
   * the pointer. If snap is false or no snap is found, push the mouse 
   * position to the snap queue. If a snap is found, push the snap.
   */
  float const * const getSnapQueue(void);

  /**
   * Wrapper function (to the marshall)
   */
  bool snapState(void);

  /**
   * Returns the current (width/height) widget ratio.
   */
  float getRatio(void);

  // Actuators // FIXME: mixed return types

  /**
   * Wrapper function (to the marshall)
   */
  bool findSnap(float, float, float);

  /**
   * Wrapper function (to the marshall)
   */
  void findIntersection(float, float, float);

  /**
   * Wrapper function (to the marshall)
   */
  unsigned int findByCoordinates(float, float, float);

  /**
   * Wrapper function (to the marshall)
   */
  void moveVertex(float, float, float);

  /**
   * Clears memory, destroys the marshall and creates a new one.
   * Nearest possible to restarting the application.
   */
  void clear(void);

  /**
   * Sets the rotation angle to zero.
   */
  void resetTransformations(void);

  // Stack operations

  /**
   * Begins an empty lines entity. Should be called at least once before
   * creating individual lines or the app will send an error.
   */
  void newLinesEntity(unsigned int layer);

  /**
   * Begins an empty triangles entity.
   */
  void newTrianglesEntity(unsigned int layer);

  /**
   * Begins a line strip. No lines are created, it only marks
   * where it begins. The layer variable determines its color and
   * on future another features too.
   */
  void newStrip(float x, float y, float z, unsigned int layer);

  /**
   * Begins a triangle strip.
   */
  void newTriStrip(float x1, float y1, float z1,
		   float x2, float y2, float z2,
		   unsigned int layer);

  /**
   * Adds a line to memory. Its color is defined by the last previous
   * call to newLinesEntity.
   */
  void pushLine(float x1, float y1, float z1,
		float x2, float y2, float z2);

  /**
   * Adds a triangle to memory.
   */
  void pushTriangle( float x1, float y1, float z1, 
		     float x2, float y2, float z2,
		     float x3, float y3, float z3);

  /**
   * Pushes a point to the active line strip, extending it by a line.
   * newStrip should have been called before the first point on the strip.
   */
  void pushToStrip(float vx, float vy, float vz);

  /**
   * Pushes a point to the active triangle strip.
   */
  void pushToTriStrip(float x1, float x2, float x3);

  /**
   * Slightly modified. Does not make current the context if it is the
   * same from where it came from.
   */
  virtual void makeCurrent();


  
  signals:

  /**
   * Emits a string with informative data about the widget.
   */
  void mousePosString(const QString &);

  /**
   * Companion to fakePythonStdout; it should be connected to a status bar.
   */
  void printStdout(const QString &);

  /**
   * Connected to a QPyObject executes the QString as a Python command
   */
  void executeCommand(const QString &);

protected:

  /**
   * Standard GL initialization.
   */
  void initializeGL();

  /**
   * Recalculates the screen ratio, keeps the center and resizes the screen 
   * in a way that top and bottom keep their values.
   */
  void resizeGL( int, int );

  /**
   * Clear, pass the matrix chain, draw.
   */
  void paintGL();
  void timerEvent( QTimerEvent * );

  /**
   * If Ctrl is pressed check x movement to modify the value of z.
   * If Btn1 is pressed check y movement to modify rotation angle.
   * Emit a string displaying the current position.
   */
  void mouseMoveEvent(QMouseEvent *); 

  /**
   * Use the mouse wheel for own needs
   */
  void wheelEvent(QWheelEvent * e);

  /**
   * Push the click to the queue. 
   * Emit a string to detail the queue state.
   */
  void mousePressEvent(QMouseEvent *);
  //  void keyPressEvent(QKeyEvent *);
};

#endif
